"use strict";
var fs$$ = require("fs"), path$$ = require("path"), resolve$$ = require("resolve"), rollup$pluginutils$$ = require("rollup-pluginutils"), $$helpers$$ = require("./helpers.js"), $$defaultResolver$$ = require("./defaultResolver.js"), $$transform$$ = require("./transform.js"), $$utils$$ = require("./utils.js");

function getCandidatesForExtension ( resolved, extension ) {
	return [
		resolved + extension,
		resolved + `${path$$.sep}index${extension}`
	];
}

function getCandidates ( resolved, extensions ) {
	return extensions.reduce(
		( paths, extension ) => paths.concat( getCandidatesForExtension ( resolved, extension ) ),
		[resolved]
	);
}

// Return the first non-falsy result from an array of
// maybe-sync, maybe-promise-returning functions
function first ( candidates ) {
	return function ( ...args ) {
		return candidates.reduce( ( promise, candidate ) => {
			return promise.then( result => result != null ?
				result :
				Promise.resolve( candidate( ...args ) ) );
		}, Promise.resolve() );
	};
}

function startsWith ( str, prefix ) {
	return str.slice( 0, prefix.length ) === prefix;
}


function commonjs(options = {}) {
	const extensions = options.extensions || ['.js'];
	const filter = rollup$pluginutils$$.createFilter( options.include, options.exclude );
	const ignoreGlobal = options.ignoreGlobal;

	const customNamedExports = {};
	if ( options.namedExports ) {
		Object.keys( options.namedExports ).forEach( id => {
			let resolvedId;

			try {
				resolvedId = resolve$$.sync( id, { basedir: process.cwd() });
			} catch ( err ) {
				resolvedId = path$$.resolve( id );
			}

			customNamedExports[ resolvedId ] = options.namedExports[ id ];
		});
	}

	const allowDynamicRequire = !!options.ignore; // TODO maybe this should be configurable?

	const ignoreRequire = typeof options.ignore === 'function' ?
		options.ignore :
		Array.isArray( options.ignore ) ? id => ~options.ignore.indexOf( id ) :
			() => false;

	let entryModuleIdPromise = null;
	let entryModuleId = null;

	function resolveId ( importee, importer ) {
		if ( importee === $$helpers$$.HELPERS_ID ) return importee;

		if ( importer && startsWith( importer, $$helpers$$.PREFIX ) ) importer = importer.slice( $$helpers$$.PREFIX.length );

		const isProxyModule = startsWith( importee, $$helpers$$.PREFIX );
		if ( isProxyModule ) importee = importee.slice( $$helpers$$.PREFIX.length );

		return resolveUsingOtherResolvers( importee, importer ).then( resolved => {
			if ( resolved ) return isProxyModule ? $$helpers$$.PREFIX + resolved : resolved;

			resolved = $$defaultResolver$$["default"]( importee, importer );

			if ( isProxyModule ) {
				if ( resolved ) return $$helpers$$.PREFIX + resolved;
				return $$helpers$$.EXTERNAL + importee; // external
			}

			return resolved;
		});
	}

	const sourceMap = options.sourceMap !== false;

	const commonjsModules = new Map();
	let resolveUsingOtherResolvers;

	return {
		name: 'commonjs',

		options ( options ) {
			const resolvers = ( options.plugins || [] )
				.map( plugin => {
					if ( plugin.resolveId === resolveId ) {
						// substitute CommonJS resolution logic
						return ( importee, importer ) => {
							if ( importee[0] !== '.' || !importer ) return; // not our problem

							const resolved = path$$.resolve( path$$.dirname( importer ), importee );
							const candidates = getCandidates( resolved, extensions );

							for ( let i = 0; i < candidates.length; i += 1 ) {
								try {
									const stats = fs$$.statSync( candidates[i] );
									if ( stats.isFile() ) return candidates[i];
								} catch ( err ) { /* noop */ }
							}
						};
					}

					return plugin.resolveId;
				})
				.filter( Boolean );

			const isExternal = options.external ?
				Array.isArray( options.external ) ? id => ~options.external.indexOf( id ) :
					options.external :
				() => false;

			resolvers.unshift( id => isExternal( id ) ? false : null );

			resolveUsingOtherResolvers = first( resolvers );

			entryModuleIdPromise = resolveId( options.input || options.entry ).then( resolved => {
				entryModuleId = resolved;
			});
		},

		resolveId,

		load ( id ) {
			if ( id === $$helpers$$.HELPERS_ID ) return $$helpers$$.HELPERS;

			// generate proxy modules
			if ( startsWith( id, $$helpers$$.EXTERNAL ) ) {
				const actualId = id.slice( $$helpers$$.EXTERNAL.length );
				const name = $$utils$$.getName( actualId );

				return `import ${name} from ${JSON.stringify( actualId )}; export default ${name};`;
			}

			if ( startsWith( id, $$helpers$$.PREFIX ) ) {
				const actualId = id.slice( $$helpers$$.PREFIX.length );
				const name = $$utils$$.getName( actualId );

				return commonjsModules.has( actualId ) ?
					`import { __moduleExports } from ${JSON.stringify( actualId )}; export default __moduleExports;` :
					`import * as ${name} from ${JSON.stringify( actualId )}; export default ( ${name} && ${name}['default'] ) || ${name};`;
			}
		},

		transform ( code, id ) {
			if ( !filter( id ) ) return null;
			if ( extensions.indexOf( path$$.extname( id ) ) === -1 ) return null;

			return entryModuleIdPromise.then( () => {
				const transformed = $$transform$$["default"]( code, id, id === entryModuleId, ignoreGlobal, ignoreRequire, customNamedExports[ id ], sourceMap, allowDynamicRequire );

				if ( transformed ) {
					commonjsModules.set( id, true );
					return transformed;
				}
			});
		}
	};
}

module.exports = commonjs;